#include "main.h"

// -----------------------------------------------------------------------------
// ОБЩИЕ МЕТОДЫ
// -----------------------------------------------------------------------------

Main::Main(int w, int h, int scale, int fps) {
    create(w, h, scale, "SDL2 Window", fps);
}

Main::~Main() {
    destroy();
}

// Создать новое окно
int Main::create(int w, int h, int scale, const char* name = "SDL2", int fps = 50) {

    unsigned format = SDL_PIXELFORMAT_BGRA32;

    _scale  = scale;
    _width  = w;
    _height = h;
    frame_id = 0;
    vga_hs   = 1;
    vga_vs   = 0;
    vga_x    = 0;
    vga_y    = 2;

    width  = w * scale;
    height = h * scale;

    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO)) {
        exit(1);
    }

    SDL_ClearError();

    // Создать окно
    sdl_window = SDL_CreateWindow(name, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, width, height, SDL_WINDOW_SHOWN);
    if (sdl_window == NULL) exit(1);

    // Создать отрисовщик текстур
    sdl_renderer = SDL_CreateRenderer(sdl_window, -1, SDL_RENDERER_PRESENTVSYNC);
    if (!sdl_renderer) exit(1);

    // Формат пикселей
    sdl_pixel_format = SDL_AllocFormat(format);
    if (!sdl_pixel_format) exit(1);

    //  Создать текстуру в памяти
    sdl_screen_texture = SDL_CreateTexture(sdl_renderer, format, SDL_TEXTUREACCESS_STREAMING, width, height);
    if (!sdl_screen_texture) exit(1);

    // Смешивания текстур нет
    SDL_SetTextureBlendMode(sdl_screen_texture, SDL_BLENDMODE_NONE);

    // Буфер экрана в памяти
    screen_buffer    = (Uint32*)malloc(width * height * sizeof(Uint32));

    // Настройка FPS
    frame_length     = 1000 / (fps ? fps : 1);
    frame_prev_ticks = SDL_GetTicks();

    return 0;
}

// Ожидание событий
int Main::event() {

    for (;;) {

        Uint32 ticks = SDL_GetTicks();

        // Ожидать наступления события
        if (SDL_PollEvent(& evt)) {

            switch (evt.type) {

                // Выход из программы по нажатии "крестика"
                case SDL_QUIT: {
                    return 0;
                }

                // https://wiki.machinesdl.org/SDL_Scancode

                // Нажатие на клавишу
                case SDL_KEYDOWN: {

                    kb = evt.key.keysym.scancode;
                    return EvtKbDown;
                }

                // Клавиша отпущена
                case SDL_KEYUP: {

                    kb = evt.key.keysym.scancode;
                    return EvtKbUp;
                }

                // Движение мыши
                case SDL_MOUSEMOTION: {

                    mx = evt.motion.x;
                    my = evt.motion.y;
                    return EvtMsMotion;
                }

                // Движение мыши
                case SDL_MOUSEBUTTONDOWN:
                case SDL_MOUSEBUTTONUP: {

                    // SDL_BUTTON_LEFT | SDL_BUTTON_MIDDLE | SDL_BUTTON_RIGHT
                    mb = evt.button.button;

                    // SDL_PRESSED | SDL_RELEASED
                    ms = evt.button.state;

                    return EvtMsButton;
                }

                // Все другие события
                default: {
                    return EvtOther;
                }
            }
        }

        // Истечение таймаута: обновление экрана
        if (ticks - frame_prev_ticks >= frame_length) {

            frame_prev_ticks = ticks;
            update();
            return EvtRedraw;
        }

        SDL_Delay(1);
    }
}

// Обновить окно
void Main::update() {

    SDL_Rect dstRect;

    dstRect.x = 0;
    dstRect.y = 0;
    dstRect.w = width;
    dstRect.h = height;

    SDL_UpdateTexture       (sdl_screen_texture, NULL, screen_buffer, width * sizeof(Uint32));
    SDL_SetRenderDrawColor  (sdl_renderer, 0, 0, 0, 0);
    SDL_RenderClear         (sdl_renderer);
    SDL_RenderCopy          (sdl_renderer, sdl_screen_texture, NULL, &dstRect);
    SDL_RenderPresent       (sdl_renderer);
}

// Уничтожение окна
void Main::destroy() {

    if (sdl_screen_texture) {
        SDL_DestroyTexture(sdl_screen_texture);
        sdl_screen_texture = NULL;
    }

    if (sdl_pixel_format) {
        SDL_FreeFormat(sdl_pixel_format);
        sdl_pixel_format = NULL;
    }

    if (sdl_renderer) {
        SDL_DestroyRenderer(sdl_renderer);
        sdl_renderer = NULL;
    }

    free(screen_buffer);

    SDL_DestroyWindow(sdl_window);
    SDL_Quit();
}

// -----------------------------------------------------------------------------
// ФУНКЦИИ РИСОВАНИЯ
// -----------------------------------------------------------------------------

// Установка точки
void Main::pset(int x, int y, Uint32 cl) {

    if (x < 0 || y < 0 || x >= _width || y >= _height)
        return;

    if (_scale == 1) {
        screen_buffer[y*width + x] = cl;
    } else {
        for (int i = 0; i < _scale; i++)
        for (int j = 0; j < _scale; j++)
            screen_buffer[(_scale*y+i)*width + (_scale*x + j)] = cl;
    }
}

// Отслеживание сигнала RGB по HS/VS; save=1 сохранить фрейм как ppm, 2=как png
void Main::vga(int R, int G, int B, int HS, int VS, int save) {

    // Отслеживание на фронтах HS/VS
    if (HS) vga_x++;

    if (vga_hs == 1 && HS == 0) { vga_x = 0; vga_y++; }
    if (vga_vs == 0 && VS == 1) { vga_y = 0; if (save) saveframe("temp", save); }

    vga_hs = HS;
    vga_vs = VS;

    // Вывод на экран
    pset(vga_x-(96-48+2), vga_y-(35-2+4), (R<<(16)) + (G<<(8)) + (B));
}

// Сохранение фрейма
void Main::saveframe(const char* dir, int type) {

    char fn[256];

    if (type == 1) {
        sprintf(fn, "%s/%08d.ppm", dir, frame_id);
    } else if (type == 2) {
        sprintf(fn, "%s/%08d.png", dir, frame_id);
    } else {
        return;
    }

    FILE* fp = fopen(fn, "wb");
    if (fp) {

        // Сохранить как PPM
        if (type == 1) {

            fprintf(fp, "P6\n# Verilator\n640 400\n255\n");
            for (int y = 0; y < 400; y++)
            for (int x = 0; x < 640; x++) {

                int cl = screen_buffer[2*(y*width + x)];
                int vl = ((cl >> 16) & 255) + (cl & 0xFF00) + ((cl&255)<<16);
                fwrite(&vl, 1, 3, fp);
            }

            fclose(fp);
        }
        // Сохранить как PNG
        else if (type == 2) {

            fclose(fp);
            unsigned char* image = (unsigned char*)malloc(4*_width*_height);

            for (int y = 0; y < _height; y++)
            for (int x = 0; x < _width; x++) {

                int cl = screen_buffer[2*(y*width + x)];

                image[4*_width*y + 4*x + 0] = cl;
                image[4*_width*y + 4*x + 1] = cl>>8;
                image[4*_width*y + 4*x + 2] = cl>>16;
                image[4*_width*y + 4*x + 3] = 255;
            }

            unsigned error = lodepng_encode32_file(fn, image, _width, _height);
            if (error) printf("error %u: %s\n", error, lodepng_error_text(error));

            free(image);
        }
    }

    frame_id++;
}

